/*
 * Copyright (C) 2000  Florian Sander
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

static void incrstats(char *user, char *chan, int type, int value, int set)
{
  globstats *gs, *gs2;
  locstats *ls, *ls2;
  int i;

  Context;
  if (type >= TOTAL_TYPES)
    return;
  if (!user) {
    debug0("Stats.mod: incrstats(..) Ups, user is NULL!");
    return;
  }
  if (!chan) {
    debug0("Stats.mod: incrstats(..) Ups, chan is NULL!");
    return;
  }
  for (gs = sdata; gs; gs = gs->next) {
    if (!strcasecmp(chan, gs->chan))
      break;
  }
  if (!gs) {
    gs2 = sdata;
    while (gs2 && gs2->next)
      gs2 = gs2->next;
    gs = nmalloc(sizeof(globstats));
    gs->started = now;
    gs->peak[S_TOTAL] = gs->peak[S_DAILY] = gs->peak[S_WEEKLY] = gs->peak[S_MONTHLY] = 0;
    for (i = 0; i < 24; i++) {
      gs->users[S_USERSUM][i] = 0;
      gs->users[S_USERCOUNTS][i] = -1;
    }
    gs->next = NULL;
    gs->local = NULL;
    gs->words = NULL;
    gs->topics = NULL;
    gs->hosts = NULL;
    gs->chan = nmalloc(strlen(chan) + 1);
    strcpy(gs->chan, chan);
    if (gs2)
      gs2->next = gs;
    else
      sdata = gs;
  }
  for (ls = gs->local; ls; ls = ls->next) {
    if (!strcasecmp(ls->user, user))
      break;
  }
  if (type == T_GSTARTED) {
    gs->started = value;
    return;
  }
  if (type == T_PEAK) {
    gs->peak[set] = value;
    return;
  }
  if (!ls) {
    ls2 = gs->local;
    while (ls2 && ls2->next)
      ls2 = ls2->next;
    ls = nmalloc(sizeof(locstats));
    ls->started = now;
    ls->next = NULL;
    ls->words = NULL;
    ls->quotes = NULL;
    ls->quotefr = 0;
    for (i = 0; i < TOTAL_TYPES; i++) {
      ls->values[S_TOTAL][i] = 0;
      ls->values[S_TODAY][i] = 0;
      ls->values[S_WEEKLY][i] = 0;
      ls->values[S_MONTHLY][i] = 0;
    }
    ls->user = nmalloc(strlen(user) + 1);
    strcpy(ls->user, user);
    if (ls2)
      ls2->next = ls;
    else
      gs->local = ls;
  }
  if (type == T_LSTARTED)
    ls->started = value;
  else {
    if (set > 0)
      ls->values[set - 1][type] = value;
    else if (set < 0)
      ls->values[(set * (-1)) - 1][type] += value;
    else {
      ls->values[S_TOTAL][type] += value;
      ls->values[S_TODAY][type] += value;
      ls->values[S_WEEKLY][type] += value;
      ls->values[S_MONTHLY][type] += value;
    }
  }
}

static int getstats(char *user, char *chan, char *type, int today)
{
  struct stats_global *gs = sdata;
  struct stats_local *ls;
  int itype;

  Context;
  while (gs) {
    ls = gs->local;
    if (!strcasecmp(gs->chan, chan)) {
      while (ls) {
        if (!strcasecmp(ls->user, user)) {
          itype = typetoi(type);
          if (itype >= 0)
            return ls->values[today][itype];
        }
        ls = ls->next;
      }
    }
    gs = gs->next;
  }
  Context;
  return 0;
}

static void sortstats(struct stats_global *gs, char *type, int today)
{
  int again = 1;
  struct stats_local *last, *p, *c, *n;
  int a, b, itype;

  Context;
  again = 1;
  last = NULL;
  itype = typetoi(type);
  while ((gs->local != last) && (again)) {
    p = NULL;
    c = gs->local;
    n = c->next;
    again = 0;
    while (n != last) {
      if (!c || !n)
        a = b = 0;
      else if (itype == T_LSTARTED) {
        a = c->started;
        b = n->started;
      } else if (itype >= 0) {
        a = c->values[today][itype];
        b = n->values[today][itype];
      } else if (itype == T_WPL) {
	if (c->values[today][T_LINES])
          a = (int) (((float) c->values[today][T_WORDS] / (float) c->values[today][T_LINES]) * 100.0);
        else
          a = 0;
        if (n->values[today][T_LINES])
          b = (int) (((float) n->values[today][T_WORDS] / (float) n->values[today][T_LINES]) * 100.0);
        else
          b = 0;
      } else if (itype == T_WORD) {
        if (c->word)
          a = c->word->nr;
        else
          a = 0;
        if (n->word)
          b = n->word->nr;
        else
          b = 0;
      } else
        a = b = 0;
      if (a < b) {
        again = 1;
        c->next = n->next;
        n->next = c;
        if (p == NULL)
          gs->local = n;
        else
          p->next = n;
      }
      p = c;
      c = n;
      n = n->next;
    }
    last = c;
  }
  Context;
  return;
}

static void sortwordstats(locstats *ls, globstats *gs)
{
  int again = 1;
  wordstats *last, *p, *c, *n, *tmp;
  int a, b;

  Context;
  again = 1;
  last = NULL;
  if (ls)
    tmp = ls->words;
  else
    tmp = gs->words;
  while ((tmp != last) && (again)) {
    p = NULL;
    if (ls)
      c = ls->words;
    else
      c = gs->words;
    n = c->next;
    again = 0;
    while (n != last) {
      if (!c || !n)
        a = b = 0;
      else {
        a = c->nr;
        b = n->nr;
      }
      if (a < b) {
  again = 1;
  c->next = n->next;
  n->next = c;
  if (p == NULL) {
    if (ls)
      ls->words = n;
    else
      gs->words = n;
    tmp = n;
  } else
    p->next = n;
      }
      p = c;
      c = n;
      n = n->next;
    }
    last = c;
  }
  Context;
  return;
}

static void sorthosts(struct stats_global *gs)
{
  int again = 1;
  hoststr *last, *p, *c, *n;
  int a, b;

  Context;
  again = 1;
  last = NULL;
  while ((gs->hosts != last) && (again)) {
    p = NULL;
    c = gs->hosts;
    n = c->next;
    again = 0;
    while (n != last) {
      if (!c || !n)
        a = b = 0;
      else {
        a = c->nr;
        b = n->nr;
      }
      if (a < b) {
        again = 1;
        c->next = n->next;
        n->next = c;
        if (p == NULL)
          gs->hosts = n;
        else
          p->next = n;
      }
      p = c;
      c = n;
      n = n->next;
    }
    last = c;
  }
  Context;
  return;
}

static int typetoi(char *type)
{
  Context;
  if (!strcasecmp(type, "lstarted"))
    return T_LSTARTED;
  else if (!strcasecmp(type, "gstarted"))
    return T_GSTARTED;
  else if (!strcasecmp(type, "words"))
    return T_WORDS;
  else if (!strcasecmp(type, "letters"))
    return T_LETTERS;
  else if (!strcasecmp(type, "minutes"))
    return T_MINUTES;
  else if (!strcasecmp(type, "topics"))
    return T_TOPICS;
  else if (!strcasecmp(type, "lines"))
    return T_LINES;
  else if (!strcasecmp(type, "actions"))
    return T_ACTIONS;
  else if (!strcasecmp(type, "kicks"))
    return T_KICKS;
  else if (!strcasecmp(type, "modes"))
    return T_MODES;
  else if (!strcasecmp(type, "bans"))
    return T_BANS;
  else if (!strcasecmp(type, "nicks"))
    return T_NICKS;
  else if (!strcasecmp(type, "joins"))
    return T_JOINS;
  else if (!strcasecmp(type, "smileys"))
    return T_SMILEYS;
  else if (!strcasecmp(type, "questions"))
    return T_QUESTIONS;
  else if (!strcasecmp(type, "wpl"))
    return T_WPL;
  else if (!strcasecmp(type, "w/l"))
    return T_WPL;
  else if (!strcasecmp(type, "word"))
    return T_WORD;
  else if (!strcasecmp(type, "started"))
    return T_LSTARTED;
  else {
    debug1("Stats.mod: Unknown stat type: %s", type);
    return T_ERROR;
  }
}

static locstats *findlocstats(char *chan, char *user)
{
  globstats *gl;
  locstats *ll;

  Context;
  for (gl = sdata; gl; gl = gl->next) {
    if (!strcasecmp(gl->chan, chan))
      break;
  }
  if (!gl)
    return NULL;
  for (ll = gl->local; ll; ll = ll->next) {
    if (!strcasecmp(ll->user, user))
      return ll;
  }
  return NULL;
}

static globstats *findglobstats(char *chan)
{
  globstats *gl;

  Context;
  for (gl = sdata; gl; gl = gl->next) {
    if (!strcasecmp(gl->chan, chan))
      break;
  }
  return gl;
}

static void write_stats()
{
  char s[125];
  FILE *f;
  struct stats_global *gs;
  struct stats_local *ls;
  int i;

  Context;
  if (!statsfile[0])
    return;
  sprintf(s, "%s~new", statsfile);
  f = fopen(s, "w");
  chmod(s, statsfilemode);
  if (f == NULL) {
    putlog(LOG_MISC, "*", "ERROR writing stats file.");
    return;
  }
  fprintf(f, "@ # Statistics from %s.\n", botnetnick);
  fprintf(f, "@ filever 1\n");
  fprintf(f, "@ month %d\n", getmonth());
  for (gs = sdata; gs; gs = gs->next) {
    fprintf(f, "%s ! %d\n", gs->chan, (int) gs->started);
    fprintf(f, "%s @ %d\n", gs->chan, gs->peak[S_TOTAL]);
    fprintf(f, "@ peaks %s %d %d %d %d\n", gs->chan, gs->peak[S_TOTAL],
            gs->peak[S_DAILY], gs->peak[S_WEEKLY], gs->peak[S_MONTHLY]);
    for (ls = gs->local; ls; ls = ls->next) {
      fprintf(f, "%s %s %d", gs->chan, ls->user, (int) ls->started);
      for (i = 0; i < TOTAL_TYPES; i++)
        fprintf(f, " %d", ls->values[S_TOTAL][i]);
      fprintf(f, "\n");
      fprintf(f, "@ daily %s %s", gs->chan, ls->user);
      for (i = 0; i < TOTAL_TYPES; i++)
        fprintf(f, " %d", ls->values[S_DAILY][i]);
      fprintf(f, "\n");
      fprintf(f, "@ weekly %s %s", gs->chan, ls->user);
      for (i = 0; i < TOTAL_TYPES; i++)
        fprintf(f, " %d", ls->values[S_WEEKLY][i]);
      fprintf(f, "\n");
      fprintf(f, "@ monthly %s %s", gs->chan, ls->user);
      for (i = 0; i < TOTAL_TYPES; i++)
        fprintf(f, " %d", ls->values[S_MONTHLY][i]);
      fprintf(f, "\n");
      i = 0;
    }
  }
  fclose(f);
  unlink(statsfile);
  movefile(s, statsfile);
  Context;
  return;
}

static void read_stats()
{
  FILE *f;
  char buf[SAVESTATSLENGTH + 1];
  char *s, *chan, *user, *cmd;
  int i, version, range, month;

  Context;
  version = 0;
  month = 0;
  f = fopen(statsfile, "r");
  if (f == NULL) {
    putlog(LOG_MISC, "*", "ERROR reading stats file");
    return;
  }
  free_stats();
  while (!feof(f)) {
    buf[0] = 0;
    s = buf;
    fgets(s, SAVESTATSLENGTH, f);
    if (buf[0] == 0)
      continue;
    if (s[strlen(s) - 1] == '\n')
      s[strlen(s) - 1] = 0;
    chan = newsplit(&s);
    if (!strcmp(chan, "@")) {
      cmd = newsplit(&s);
      if (!strcmp(cmd, "filever"))
	version = atoi(newsplit(&s));
      else if (!strcmp(cmd, "month"))
        month = atoi(newsplit(&s));
      else if (!strcmp(cmd, "peaks")) {
        chan = newsplit(&s);
        incrstats("*", chan, T_PEAK, atoi(newsplit(&s)), S_TOTAL);
        incrstats("*", chan, T_PEAK, atoi(newsplit(&s)), S_DAILY);
        incrstats("*", chan, T_PEAK, atoi(newsplit(&s)), S_WEEKLY);
        incrstats("*", chan, T_PEAK, atoi(newsplit(&s)), S_MONTHLY);
      } else if (!strcmp(cmd, "daily") || !strcmp(cmd, "weekly")
                 || !strcmp(cmd, "monthly")) {
        if (!strcmp(cmd, "daily"))
	  range = S_DAILY;
        else if (!strcmp(cmd, "weekly"))
	  range = S_WEEKLY;
        else if (!strcmp(cmd, "monthly"))
          range = S_MONTHLY;
        else {
          debug2("Error while reading statsfile: range uninitialized! (%s %s)", cmd, s);
          continue;
        }
        if ((range == S_MONTHLY) && (month != lastmonth))
          continue;
        chan = newsplit(&s);
        user = newsplit(&s);
        for (i = 0; i < TOTAL_TYPES; i++)
          incrstats(user, chan, i, atoi(newsplit(&s)), range + 1);
      }
    } else {
      user = newsplit(&s);
      if (!strcmp(user, "!"))
        incrstats(user, chan, T_GSTARTED, atoi(newsplit(&s)), 1);
      else if (!strcmp(user, "@"))
        incrstats(user, chan, T_PEAK, atoi(newsplit(&s)), S_TOTAL);
      else {
        incrstats(user, chan, T_LSTARTED, atoi(newsplit(&s)), 1);
        for (i=0; i < TOTAL_TYPES; i++)
          incrstats(user, chan, i, atoi(newsplit(&s)), 1);
      }
    }
  }
  fclose(f);
  Context;
  return;
}

static void reset_tstats()
{
  globstats *gs;
  locstats *ls;
  int i;

  Context;
  putlog(LOG_MISC, "*", "Stats.mod: Resetting today's statistics...");
  for (gs = sdata; gs; gs = gs->next) {
    gs->peak[S_TODAY] = 0;
    free_wordstats(gs->words);
    gs->words = NULL;
    free_topics(gs->topics);
    gs->topics = NULL;
    free_hosts(gs->hosts);
    gs->hosts = NULL;
    for (ls = gs->local; ls; ls = ls->next) {
      free_wordstats(ls->words);
      ls->words = NULL;
      free_quotes(ls->quotes);
      ls->quotes = NULL;
      for (i = 0; i < TOTAL_TYPES; i++)
        ls->values[S_TODAY][i] = 0;
    }
  }
  Context;
}

static void reset_mwstats(int range)
{
  globstats *gs;
  locstats *ls;
  int i;

  Context;
  putlog(LOG_MISC, "*", "Stats.mod: Resetting %s statistics...", (range == S_WEEKLY) ? "weekly" : "monthly");
  for (gs = sdata; gs; gs = gs->next) {
    for (ls = gs->local; ls; ls = ls->next) {
      for (i = 0; i < TOTAL_TYPES; i++)
        ls->values[range][i] = 0;
    }
  }
  Context;
}

static void sort_stats_alphabetically(globstats *gs)
{
  locstats *as, *bs, *l, *last;
  int a, b, again = 1;
  char *astr, *bstr, n[2];

  Context;
  n[0] = n[1] = 0;
  last = NULL;
  while ((gs->local != last) && again) {
    again = 0;
    l = NULL;
    as = gs->local;
    bs = gs->local->next;
    while(bs) {
      if (!as)
        astr = n;
      else
        astr = as->user;
      if (!bs)
        bstr = n;
      else
        bstr = bs->user;
      a = (int) tolower(astr[0]);
      b = (int) tolower(bstr[0]);
      while ((a == b) && a && b) {
	astr++;
	bstr++;
        a = (int) tolower(astr[0]);
        b = (int) tolower(bstr[0]);
      }
      if (a > b) {
        if (!l)
          gs->local = bs;
        else
          l->next = bs;
        as->next = bs->next;
        bs->next = as;
        again = 1;
        if (l == NULL)
          gs->local = bs;
        else
          l = bs;
      }
      l = as;
      as = bs;
      bs = bs->next;
    }
    last = as;
  }
  Context;
}

static void resetlocstats(locstats *ls)
{
  int i;

  if (!ls) {
    debug0("ERROR! resetlocstats called with NULL pointer!");
    return;
  }
  for (i = 0; i < TOTAL_TYPES; i++) {
    ls->values[S_TOTAL][i] = 0;
    ls->values[S_TODAY][i] = 0;
  }
  return;
}

static void calcwordstats(char *hand, char *chan, char *rest)
{
  locstats *ls;
  char *word;
  int i;

  Context;
  if (!log_wordstats)
    return;
  ls = findlocstats(chan, hand);
  if (!ls) {
    debug2("Can't calculate wordstats for %s in %s, no locstats.", hand, chan);
    return;
  }
  for (i = 0; i < strlen(rest); i++)
    if (strchr("!?.,\"<>&\\", rest[i]))
      rest[i] = ' ';
  while (rest[0]) {
    word = newsplit(&rest);
    strlower(word);
    incrwordstats(hand, chan, word, 1, 0);
  }
}

static void incrwordstats(char *hand, char *chan, char *word, int value, int set)
{
  locstats *ls;
  wordstats *l, *ll;

  Context;
  if ((word[0] == ' ') || !word[0])
    return;
  ls = findlocstats(chan, hand);
  if (!ls) {
    debug2("Can't calculate wordstats for %s in %s, no locstats.", hand, chan);
    return;
  }
  for (l = ls->words; l; l = l->next)
    if (!strcmp(word, l->word))
      break;
  if (!l) {
    l = ls->words;
    while (l && l->next)
      l = l->next;
    ll = nmalloc(sizeof(wordstats));
    ll->word = nmalloc(strlen(word) + 1);
    strcpy(ll->word, word);
    ll->nr = 0;
    ll->next = NULL;
    if (l)
      l->next = ll;
    else
      ls->words = ll;
    l = ll;
  }
  if (set)
    l->nr = value;
  else
    l->nr += value;
}

static void nincrwordstats(globstats *gs, char *word, int value)
{
  wordstats *l, *ll;

  Context;
  for (l = gs->words; l; l = l->next)
    if (!strcmp(word, l->word))
      break;
  if (!l) {
    l = gs->words;
    while (l && l->next)
      l = l->next;
    ll = nmalloc(sizeof(wordstats));
    ll->word = nmalloc(strlen(word) + 1);
    strcpy(ll->word, word);
    ll->nr = 0;
    ll->next = NULL;
    if (l)
      l->next = ll;
    else
      gs->words = ll;
    l = ll;
  }
  l->nr += value;
}

static void do_globwordstats(globstats *gs)
{
  wordstats *l;
  locstats *ls;

  Context;
  for (l = gs->words; l; l = l->next)
    l->nr = 0;
  for (ls = gs->local; ls; ls = ls->next)
    for (l = ls->words; l; l = l->next)
      nincrwordstats(gs, l->word, l->nr);
  sortwordstats(NULL, gs);
}

/*
static void purgewordstats()
{
  globstats *gs;
  locstats *ls;
  wordstats *l;
  int i;

  Context;
  for (gs = sdata; gs; gs = gs->next) {
    for (ls = gs->local; ls; ls = ls->next) {
      i = 0;
      for (l = ls->words; l; l = l->next) {
        i++;
        if (i > max_words) {
          free_wordstats(l->next);
          l->next = NULL;
        }
      }
    }
  }
}
*/

static void addquote(char *user, char *chan, char *quote)
{
  quotestr *l, *nl;
  locstats *ls;

  Context;
  ls = findlocstats(chan, user);
  if (!ls) {
    debug2("Can't calculate wordstats for %s in %s, no locstats.", user, chan);
    return;
  }
  ls->quotefr--;
  if (ls->quotefr > 0)
    return;
  ls->quotefr = quote_freq;
  l = ls->quotes;
  while (l && l->next)
    l = l->next;
  nl = nmalloc(sizeof(quotestr));
  nl->next = NULL;
  nl->quote = nmalloc(strlen(quote) + 1);
  strcpy(nl->quote, quote);
  if (l)
    l->next = nl;
  else
    ls->quotes = nl;
}

static void addtopic(char *channel, char *topic, char *by)
{
  topicstr *e, *ne;
  globstats *gs;

  Context;
  gs = findglobstats(channel);
  if (!gs)
    return;
  for (e = gs->topics; e; e = e->next)
    if (!strcasecmp(topic, e->topic))
      return;
  e = gs->topics;
  while (e && e->next)
    e = e->next;
  ne = nmalloc(sizeof(topicstr));
  ne->topic = nmalloc(strlen(topic) + 1);
  strcpy(ne->topic, topic);
  ne->by = nmalloc(strlen(by) + 1);
  strcpy(ne->by, by);
  ne->when = now;
  ne->next = NULL;
  if (e)
    e->next = ne;
  else
    gs->topics = ne;
}

static void addhost(char *host, globstats *gs)
{
  hoststr *e, *ne;
  char *s;

  Context;
  if (!gs || !host)
    return;
  s = strchr(host, '@');
  if (s)
    host = s + 1;
  if (strcmp(host, "[IP]"))
    strlower(host);
  for (e = gs->hosts; e; e = e->next) {
    if (!strcmp(host, e->host)) {
      e->nr++;
      return;
    }
  }
  e = gs->hosts;
  while (e && e->next)
    e = e->next;
  ne = nmalloc(sizeof(hoststr));
  ne->host = nmalloc(strlen(host) + 1);
  strcpy(ne->host, host);
  ne->nr = 1;
  ne->next = NULL;
  if (e)
    e->next = ne;
  else
    gs->hosts = ne;
  return;
}

static void setword(globstats *gs, char *word)
{
  locstats *l;
  wordstats *w;

  for (l = gs->local; l; l = l->next) {
    l->word = NULL;
    for (w = l->words; w; w = w->next) {
      if (!strcmp(w->word, word)) {
        l->word = w;
        break;
      }
    }
  }
}
