static struct dcc_table LIVESTATS_LISTEN =
{
  "LIVESTATS_LISTEN",
  DCT_VALIDIDX,
  eof_livestats,
  livestats_accept,
  0,
  timeout_listen_livestats,
  display_livestats_accept,
  0,
  NULL,
  0
};

static struct dcc_table LIVESTATS =
{
  "LIVESTATS",
  DCT_VALIDIDX,
  eof_livestats,
  livestats_activity,
  &livestats_timeout,
  timeout_livestats,
  display_livestats,
  expmem_livestats,
  kill_livestats,
#ifdef OLDBOT
  out_livestats,
#else
  out_livestats,
  outdone_livestats
#endif
};

static int inactivechan(char *chan)
{
  struct chanset_t *ch;

  ch = findchan_by_dname(chan);
  if (!ch)
    return 1;
  if (ch->status & CHAN_INACTIVE)
    return 1;
  return 0;
}

static void start_listen_livestats(int port)
{
  int i, zz;

  stop_listen_livestats();
  zz = open_listen(&port);
  if (zz == (-1)) {
    putlog(LOG_MISC, "*", "ERROR! Cannot open listening socket for livestats!");
    return;
  }
  if ((i = new_dcc(&LIVESTATS_LISTEN, 0)) == -1) {
    putlog(LOG_MISC, "*", "ERROR! Cannot open listening socket for livestats! DCC table is full!");
    return;
  }
  dcc[i].sock = zz;
  dcc[i].addr = (IP) (-559026163);
  dcc[i].port = port;
  strcpy(dcc[i].nick, "livestats");
  strcpy(dcc[i].host, "*");
  dcc[i].timeval = now;
  putlog(LOG_MISC, "*", "Now listening for livestats connections on port %d", port);
}

static void stop_listen_livestats()
{
  int i;

  for (i = 0; i < dcc_total; i++) {
    if (dcc[i].type == &LIVESTATS_LISTEN) {
      putlog(LOG_MISC, "*",
      	     "no longer listening for livestats connections on port %d",
             dcc[i].port);
      killsock(dcc[i].sock);
      lostdcc(i);
    } else if (dcc[i].type == &LIVESTATS) {
      putlog(LOG_MISC, "*", "killing livestats connection from %s", dcc[i].host);
      killsock(dcc[i].sock);
      lostdcc(i);
    }
  }
}

static void livestats_activity(int idx, char *buf, int len)
{
  char *cmd, *path, *newpath;
  int lev;

  if (!strncasecmp(buf, "GET ", 4)) {
    if (!stats_info_access(idx)->cmd) {
      stats_info_access(idx)->cmd = nmalloc(strlen(buf) + 1);
      strcpy(stats_info_access(idx)->cmd, buf);
    }
  }
  cmd = newsplit(&buf);
  if (!strcmp(cmd, "GET")) {
    lev = logmodes(stats_loglevel);
    if (lev)
      putlog(lev, "*", "%s: GET %s", dcc[idx].host, buf);
    path = newsplit(&buf);
    if (!strncasecmp(path, "/robots.txt", 11)) {
      dprintf(idx, "HTTP/1.1 404 Not Found\nServer: stats.mod/%s\nConnection: close", MODULE_VERSION);
      dprintf(idx, "Content-Type: text/html; charset=iso-8859-1\n\n");
      dprintf(idx, "<HTML><HEAD><TITLE>404 Not Found</TITLE></HEAD><BODY><H1>Not Found</H1>");
      dprintf(idx, "The requested document was not found on this server.</BODY></HTML>");
      stats_info_access(idx)->code = 404;
      return;
    }
    if (path[strlen(path) - 1] != '/') {
      newpath = nmalloc(strlen(path) + 2);
      strcpy(newpath, path);
      newpath[strlen(path)] = '/';
      newpath[strlen(path) + 1] = 0;
      dprintf(idx, "HTTP/1.1 301 Moved Permanently\nServer: stats.mod/%s\n", MODULE_VERSION);
      dprintf(idx, "Location: %s\nConnection: close\nContent-Type: text/html\n\n", newpath);
      dprintf(idx, "<HTML><body>The concluding \"/\" is important!<br><center>");
      dprintf(idx, "<a href=\"%s\">%s</a></center><br>", newpath, newpath);
      stats_info_access(idx)->code = 301;
      nfree(newpath);
      return;
    }
    dprintf(idx, "HTTP/1.0 200 OK\nServer: stats.mod/%s\nContent-Type: text/html\n\n", MODULE_VERSION);
    send_livestats(idx, path);
  } else if (!strcasecmp(cmd, "User-Agent:")) {
    if (stats_info_access(idx)->browser)
      return;
    stats_info_access(idx)->browser = nmalloc(strlen(buf) + 1);
    strcpy(stats_info_access(idx)->browser, buf);
  } else if (!strcasecmp(cmd, "Referer:")) {
    if (stats_info_access(idx)->referer)
      return;
    stats_info_access(idx)->referer = nmalloc(strlen(buf) + 1);
    strcpy(stats_info_access(idx)->referer, buf);
  } else if (!buf[0]) {
    dcc[idx].status = 1;
#ifndef OLDBOT
    /* If there's no data in our socket, we immediately get rid of it.
     */
    if (!sock_has_data(SOCK_DATA_OUTGOING, dcc[idx].sock)) {
      killsock(dcc[idx].sock);
      lostdcc(idx);
    }
#endif
  }
}

#ifndef OLDBOT
static void outdone_livestats(int idx)
{
  if (dcc[idx].status) {
    killsock(dcc[idx].sock);
    lostdcc(idx);
  } else
    dcc[idx].status = 1;
}
#endif

static void display_livestats(int idx, char *buf)
{
  sprintf(buf, "livestats");
}

static void display_livestats_accept(int idx, char *buf)
{
  sprintf(buf, "lstn port");
}

static void timeout_livestats(int idx)
{
  killsock(dcc[idx].sock);
  lostdcc(idx);
}

static void timeout_listen_livestats(int idx)
{
  debug0("timeout listen");
  killsock(dcc[idx].sock);
  lostdcc(idx);
}

static void kill_livestats(int idx, void *x)
{
  register struct stats_clientinfo *p = (struct stats_clientinfo *) x;
  char ts[41], test[11];
  time_t tt;
  FILE *f;

  Context;
  tt = now;
  if (!p) {
    putlog(LOG_MISC, "*", "Can't kill clientinfo, no pointer. This should not happen!");
    return;
  }
  ctime(&tt);
  /* 05/Feb/2000:12:08:17 +0100 */
  strftime(test, 19, "%z", localtime(&tt));
  if (test[0] != 'z')
    strftime(ts, 40, "%d/%b/%Y:%H:%M:%S %z", localtime(&tt));
  else
    strftime(ts, 40, "%d/%b/%Y:%H:%M:%S", localtime(&tt));
  if (livestats_log[0]) {
    f = fopen(livestats_log, "a");
    if (f == NULL)
      putlog(LOG_MISC, "*", "ERROR writing livestats log.");
    else {
      if (test[0] != 'z')
        fprintf(f,
    	 "%s - - [%s] \"%s\" %d %d \"%s\" \"%s\"\n", dcc[idx].host, ts,
	 p->cmd ? p->cmd : "", p->code, p->traffic,
	 p->referer ? p->referer : "-", p->browser ?  p->browser : "");
      else
        fprintf(f,
    	 "%s - - [%s %+05d] \"%s\" %d %d \"%s\" \"%s\"\n", dcc[idx].host,
	 ts, offset * (-1) * 100, p->cmd   ? p->cmd : "", p->code ,p->traffic,
	 p->referer ? p->referer : "-", p->browser ?  p->browser : "");
      fclose(f);
    }
  }
  if (p->browser)
    nfree(p->browser);
  if (p->referer)
    nfree(p->referer);
  if (p->cmd)
    nfree(p->cmd);
  nfree(p);
}

static int expmem_livestats(void *x)
{
  register struct stats_clientinfo *p = (struct stats_clientinfo *) x;
  int tot = sizeof(struct stats_clientinfo);

  Context;
  if (!p) {
    putlog(LOG_MISC, "*", "Can't expmem clientinfo, no pointer. This should not happen!");
    return 0;
  }
  if (p->browser)
    tot += strlen(p->browser) + 1;
  if (p->referer)
    tot += strlen(p->referer) + 1;
  if (p->cmd)
    tot += strlen(p->cmd) + 1;
  return tot;
}

static void out_livestats(int idx, char *buf, void *x)
{
  register struct stats_clientinfo *p = (struct stats_clientinfo *) x;

  if (!p) {
    putlog(LOG_MISC, "*", "No stats_clientinfo pointer. This should not happen!");
    return;
  }
  p->traffic += strlen(buf);
  tputs(dcc[idx].sock, buf, strlen(buf));
}

static void livestats_accept(int idx, char *buf, int len)
{
  unsigned long ip;
  unsigned short port;
  int j = 0, sock, i;
  char s[UHOSTLEN];

  Context;
  if (dcc_total + 1 >= max_dcc) {
    j = answer(dcc[idx].sock, s, &ip, &port, 0);
    if (j != -1) {
      dprintf(-j, "Sorry, too many connections already.\r\n");
      killsock(j);
    }
    return;
  }
  sock = answer(dcc[idx].sock, s, &ip, &port, 0);
  if (sock < 0) {
    neterror(s);
    putlog(LOG_MISC, "*", "Stats.mod: Error accepting livestats connection: %s", s);
    return;
  }
  if ((i = new_dcc(&LIVESTATS, sizeof(struct stats_clientinfo))) == (-1)) {
    putlog(LOG_MISC, "*", "Error accepting livestats connection. DCC table is full.");
    killsock(sock);
    return;
  }
  dcc[i].sock = sock;
  dcc[i].addr = ip;
  dcc[i].port = port;
  strcpy(dcc[i].nick, "httpstats");
#ifndef OLDBOT
  sprintf(s, "%s", iptostr(my_htonl(ip)));
#else
  sprintf(s, "%lu.%lu.%lu.%lu", (ip >> 24) & 0xff, (ip >> 16) & 0xff,
  	  (ip >> 8) & 0xff, ip & 0xff); /* dw */
#endif
  strcpy(dcc[i].host, s);
  dcc[i].timeval = now;
  dcc[i].status = 0;
  ((struct stats_clientinfo *) dcc[i].u.other)->traffic = 0;
  ((struct stats_clientinfo *) dcc[i].u.other)->code = 200;
  ((struct stats_clientinfo *) dcc[i].u.other)->browser = NULL;
  ((struct stats_clientinfo *) dcc[i].u.other)->referer = NULL;
  ((struct stats_clientinfo *) dcc[i].u.other)->cmd = NULL;
}

static void eof_livestats(int idx)
{
  debug0("eof accept");
  killsock(dcc[idx].sock);
  lostdcc(idx);
}

static void send_livestats(int idx, char *buf)
{
  char *channel, *command;
  char what[512], *pwhat, *type, *stoday, *user, prefix[4], s1[121];
  struct chanset_t *chan;
  memberlist *m;
  int today = 0;
  globstats *gs;
  locstats *ls, *ls2;
  time_t tt, ttbuf;
  struct userrec *u;
  int itype, nr, i, ii;
  wordstats *ws;
  quotestr *qs;
  unsigned long x;

  ttbuf = now;
  tt = now;
  if (!strcmp(buf, "/")) {
    dprintf(idx, "<html><head><title>%s's Livestats</title></head>", botnetnick);
    dprintf(idx, "%s<br><br><br>\n", bodytag);
    dprintf(idx, "<table border=1 width=100%%>\n");
    for (gs = sdata; gs; gs = gs->next) {
      if (!inactivechan(gs->chan) && !secretchan(gs->chan)) {
	if (gs->chan[0] == '!')
	  strcpy(prefix, "!");
	else if (gs->chan[0] == '+')
	  strcpy(prefix, "+");
	else if (gs->chan[0] == '&')
	  strcpy(prefix, "&");
	else
	  prefix[0] = 0;
#ifndef OLDBOT
        chan = findchan_by_dname(gs->chan);
#else
        chan = findchan(gs->chan);
#endif
        if (chan && chan->channel.topic)
          dprintf(idx, "<tr><td rowspan=6 align=center width=50%%><a href=\"/%s%s/\">%s</a><br><font size=-1>(%s)</font></td>",
                  prefix, gs->chan + 1, gs->chan, filt2(chan->channel.topic));
        else
          dprintf(idx, "<tr><td rowspan=6 align=center width=50%%><a href=\"/%s%s/\">%s</a></td>",
                  prefix, gs->chan + 1, gs->chan);
        dprintf(idx, "<td rowspan=4 align=center>top%d</td>", webnr);
        dprintf(idx, "<td align=center><a href=\"%s%s/top/total/words/\">total</a></td>", prefix, gs->chan + 1);
        dprintf(idx, "</tr>\n<tr>");
        dprintf(idx, "<td align=center><a href=\"%s%s/top/daily/words/\">daily</a></td>", prefix, gs->chan + 1);
        dprintf(idx, "</tr>\n<tr>");
        dprintf(idx, "<td align=center><a href=\"%s%s/top/weekly/words/\">weekly</a></td>", prefix, gs->chan + 1);
        dprintf(idx, "</tr>\n<tr>");
        dprintf(idx, "<td align=center><a href=\"%s%s/top/monthly/words/\">monthly</a></td>", prefix, gs->chan + 1);
        dprintf(idx, "</tr>\n<tr>");
        dprintf(idx, "<td align=center colspan=2><a href=\"%s%s/users/\">users</a></td></tr>", prefix, gs->chan + 1);
        dprintf(idx, "<tr><td align=center colspan=2><a href=\"%s%s/misc/\">misc stats</a></td>", prefix, gs->chan + 1);
        /* dprintf(idx, "</tr></table>\n"); */
        dprintf(idx, "</tr><tr><td colspan=3 height=0><font size=-5>&nbsp;</font></td></tr>\n");
      }
    }
    dprintf(idx, "<tr><th colspan=3 align=center>");
    dprintf(idx, "<a href=\"http://www.visions-of-fantasy.de/stats.mod/\">");
    dprintf(idx, "Stats.mod v%s</a></th></tr></table>\n", MODULE_VERSION);
    dprintf(idx, "</body></html>\n");
    return;
  }
  if (buf[0] == '/') {
    if (!strncasecmp(buf, "/e/", 3)) {
      buf += 2;
      buf[0] = '!';
    } else if (!strncasecmp(buf, "/p/", 3)) {
      buf += 2;
      buf[0] = '+';
    } else if (!strncasecmp(buf, "/a/", 3)) {
      buf += 2;
      buf[0] = '&';
    } else if (buf[1] && strchr("+&!", buf[1]))
      buf++;
    else
      buf[0] = '#';
  }
  if (buf[strlen(buf) - 1] == '/')
    buf[strlen(buf) - 1] = 0;
  channel = splitpath(&buf);
#ifndef OLDBOT
  chan = findchan_by_dname(channel);
#else
  chan = findchan(channel);
#endif
  if (!chan) {
    dprintf(idx, "no such channel %s", channel);
    stats_info_access(idx)->code = 404;
    return;
  }
  if (!strcmp(buf, "/") || !buf[0]) {
    dprintf(idx, "<html><head><title>%s statistics index</title></head>%s<br><br><br><center>\n", channel, bodytag);
    dprintf(idx, "<table border=1 width=50%%>\n");
    dprintf(idx, "<tr><th colspan=2><font size=+2>%s</font></th></tr>", channel);
    dprintf(idx, "<tr><td rowspan=4 align=center>top%d</td>", webnr);
    dprintf(idx, "<td align=center><a href=\"top/total/words/\">total</a></td>");
    dprintf(idx, "</tr>\n<tr>");
    dprintf(idx, "<td align=center><a href=\"top/daily/words/\">daily</a></td>");
    dprintf(idx, "</tr>\n<tr>");
    dprintf(idx, "<td align=center><a href=\"top/weekly/words/\">weekly</a></td>");
    dprintf(idx, "</tr>\n<tr>");
    dprintf(idx, "<td align=center><a href=\"top/monthly/words/\">monthly</a></td>");
    dprintf(idx, "</tr>\n<tr>");
    dprintf(idx, "<td align=center colspan=2><a href=\"users/\">users</a></td>");
    dprintf(idx, "</tr><tr><td align=center colspan=2><a href=\"misc/\">misc stats</a></td>");
    dprintf(idx, "</tr><tr><td colspan=2 height=0><font size=-5>&nbsp;</font></td></tr>\n");
    dprintf(idx, "<tr><th colspan=2 align=center><font size=-2>");
    dprintf(idx, "<a href=\"http://www.visions-of-fantasy.de/stats.mod/\">");
    dprintf(idx, "Stats.mod v%s</a></font></th></tr></table>\n", MODULE_VERSION);
    dprintf(idx, "</center></body></html>\n");
    return;
  }
  command = splitpath(&buf);
  if (!strcasecmp(command, "top")) {
    if (!strcmp(buf, "/") || !buf[0]) {
      dprintf(idx, "<html><body><table>");
      dprintf(idx, "<tr><td><a href=\"total/\">total</a></td></tr>");
      dprintf(idx, "<tr><td><a href=\"today/\">today</a></td></tr>");
      dprintf(idx, "</table></body></html>");
      return;
    }
    stoday = splitpath(&buf);
    if (!strcasecmp(stoday, "today") || !strcasecmp(stoday, "daily"))
      today = S_DAILY;
    else if (!strcasecmp(stoday, "weekly"))
      today = S_WEEKLY;
    else if (!strcasecmp(stoday, "monthly"))
      today = S_MONTHLY;
    else if (!strcasecmp(stoday, "total"))
      today = 0;
    else {
      dprintf(idx, "<html><body>Error, invalid value %s<table>", stoday);
      dprintf(idx, "<tr><td><a href=\"../total/\">total</a></td></tr>");
      dprintf(idx, "<tr><td><a href=\"../daily/\">daily</a></td></tr>");
      dprintf(idx, "<tr><td><a href=\"../weekly/\">weekly</a></td></tr>");
      dprintf(idx, "<tr><td><a href=\"../monthly/\">monthly</a></td></tr>");
      dprintf(idx, "</table></body></html>");
      stats_info_access(idx)->code = 404;
      return;
    }
    if (!strcmp(buf, "/") || !buf[0]) {
      dprintf(idx, "<html><body><table>");
      sprintf(what, "%s", webstats);
      pwhat = what;
      while (strlen(pwhat) > 0) {
        type = newsplit(&pwhat);
        itype = typetoi(type);
        if (itype >= 0) {
          dprintf(idx, "<tr><td><a href=\"%s/\">%s</a></td></tr>", type, type);
        }
      }
      dprintf(idx, "</table></body></html>");
      return;
    }
    gs = findglobstats(channel);
    if (!gs) {
      debug1("Error! Can't find global stats for %s", channel);
      dprintf(idx, "<html><body>Error! Can't find global stats for %s</body></html>", channel);
      stats_info_access(idx)->code = 404;
      return;
    }
    if (!strncasecmp(buf, "graph", 5)) {
      do_graphs(idx, today, gs, channel);
      return;
    }
    do_toptalkers(idx, today, gs, channel, buf);
    return;
  } else if (!strcasecmp(command, "users")) {
    gs = findglobstats(channel);
    if (!gs) {
      debug1("Error! Can't find global stats for %s", channel);
      dprintf(idx, "<html><body>Error! Can't find global stats for %s</body></html>", channel);
      return;
    }
    if (!strcmp(buf, "/") || !buf[0]) {
      sort_stats_alphabetically(gs);
      dprintf(idx, "<html><head><title>users in %s</title></head>%s<center>\n",
              channel, bodytag);
      if (chan && (chan->channel.members > 0)) {
        dprintf(idx, "<br><br><table border=1>\n<tr><th colspan=3>Currently on channel:</th></tr>\n");
        dprintf(idx, "<tr><th align=\"center\">nick</th><th align=\"center\">user</th><th align=\"center\">idle time</th></tr>\n");
        for (m = chan->channel.member; m && m->nick[0]; m = m->next) {
          dprintf(idx, "<tr><td align=\"right\">%s</td>", m->nick);
          if (!m->user && ((strlen(m->nick) + strlen(m->userhost) + 1) < 120)) {
            sprintf(s1, "%s!%s", m->nick, m->userhost);
            m->user = get_user_by_host(s1);
          }
          if (m->user)
            dprintf(idx, "<td align=\"center\"><a href=\"%s/\">%s</a></td>", m->user->handle, m->user->handle);
          else
            dprintf(idx, "<td align=\"center\">-</td>");
          if (chan_issplit(m))
            dprintf(idx, "<td>netsplitted</td>");
          else if (!rfc_casecmp(m->nick, botname))
            dprintf(idx, "<td>it's me ^_^</td>");
          else
            dprintf(idx, "<td>%s</td>", duration(now - m->last));
          dprintf(idx, "</tr>\n");
        }
        dprintf(idx, "</table>\n<br><br>");
      }

      dprintf(idx, "<table border=1>\n<tr><th>User</th><th>Info</th></tr>\n");
      for (ls = gs->local; ls; ls = ls->next) {
        u = get_user_by_handle(userlist, ls->user);
        if (!u)
          continue;
        what[0] = 0;
        get_handle_chaninfo(ls->user, channel, what);
        pwhat = what;
        if (!what[0])
          pwhat = get_user(&USERENTRY_INFO, u);
        if (!pwhat) {
          sprintf(what, "&nbsp;");
          pwhat = what;
        }
        dprintf(idx, "<tr><td><a href=\"%s/\">%s</a></td><td>%s</td></tr>\n", ls->user, ls->user, filt2(pwhat));
      }
/*
*     dprintf(idx, "<tr><td colspan=2 align=center><a href=\"http://www.visions-of-fantasy.de/stats.mod/\">");
*     dprintf(idx, "Stats.mod v%s</a></td></tr></table>\n", MODULE_VERSION);
*/
      dprintf(idx, "</table>\n");
      dprintf(idx, "<br><br><hr>\n");
      dprintf(idx, "<table width=100%% border=0>\n");
      dprintf(idx, "<tr><td align=center><table width=100%% border=0><tr>\n");
      dprintf(idx, "<td width=33%% align=center><font size=-1><a href=\"../top/total/words/\">top%d</a></font></td>\n", webnr);
      dprintf(idx, "<td width=33%% align=center><font size=-1><a href=\"../misc/\">misc stats</a></font></td>\n");
      dprintf(idx, "<td width=33%% align=center><font size=-1><a href=\"/\">other channels</a></font></td>\n");
      dprintf(idx, "</tr></table></td></tr>\n");
      dprintf(idx, "</table>\n");
      dprintf(idx, "<center>Created by <a href=\"http://www.visions-of-fantasy.de/stats.mod/\">Stats.mod v%s</a></center></body></html>", MODULE_VERSION);
      dprintf(idx, "</center></body></html>");
      return;
    }
    user = buf;
    ls = findlocstats(channel, user);
    if (!ls) {
      debug1("Error! Can't find global stats for %s", channel);
      dprintf(idx, "<html><body>I don't have any stats for %s in %s, sorry</body></html>", user, channel);
      stats_info_access(idx)->code = 404;
      return;
    }
    i = countwords(webstats) + 2;
    dprintf(idx, "<html><head><meta http-equiv=\"expires\" content=\"0\">\n");
    dprintf(idx, "<title>Stats for %s in %s</title></head>%s<center>\n", ls->user, channel, bodytag);
    dprintf(idx, "<table border=1><tr><th colspan=%d align=center>Stats for %s in %s</th></tr>\n", i, ls->user, channel);
    dprintf(idx, "<tr><td colspan=%d>&nbsp;</td> </tr><tr><td></td><td>place</td>", i);
    sprintf(what, "%s", webstats);
    pwhat = what;
    while (strlen(pwhat) > 0) {
      dprintf(idx, "<td>%s</td>", newsplit(&pwhat));
    }
    dprintf(idx, "</tr>\n");
    for (today = 0; today < 4; today++) {
      dprintf(idx, "<tr><td align=center>%s</td>", RANGESTR_LONG);
      nr = 0;
      if (ls->values[today][T_WORDS] > 0) {
        sortstats(gs, "words", today);
        for (ls2 = gs->local; ls2; ls2 = ls2->next) {
          nr++;
          if (!strcasecmp(ls2->user, user))
            break;
        }
      }
      if (nr == 0)
        dprintf(idx, "<td>-</td>");
      else
        dprintf(idx, "<td>%d</td>", nr);
      sprintf(what, "%s", webstats);
      pwhat = what;
      while (strlen(pwhat) > 0) {
        itype = typetoi(newsplit(&pwhat));
        if (itype == T_MINUTES)
          dprintf(idx, "<td>%s</td>", duration(ls->values[today][itype] * 60));
        else if (itype == T_WPL) {
	  if (ls->values[today][T_LINES])
            dprintf(idx, "<td>%.2f</td>", (float) ls->values[today][T_WORDS] / (float) ls->values[today][T_LINES]);
          else
            dprintf(idx, "<td>0</td>");
        } else if (itype >= 0)
          dprintf(idx, "<td>%d</td>", ls->values[today][itype]);
        else
          dprintf(idx, "<td>ERR!</td>");
      }
      dprintf(idx, "</tr>\n");
    }
    if (ls->quotes) {
      nr = 0;
      for (qs = ls->quotes; qs; qs = qs->next)
        nr++;
      x = random() % nr;
      ii = 0;
      qs = ls->quotes;
      while (qs) {
        if (ii == x) {
          dprintf(idx, "<tr><td colspan=%d align=center>random quote: \"%s\"</td></tr>\n", i, filt2(qs->quote));
          break;
        }
        qs = qs->next;
        ii++;
      }
    }
    /* dprintf(idx, "<tr><td colspan=%d align=center>\
    	    <a href=\"http://www.visions-of-fantasy.de/stats.mod/\">\
	    Stats.mod v%s</a></td></tr>\n", i, MODULE_VERSION); */
    dprintf(idx, "</table></center>\n");
    if (ls->words || ls->quotes) {
      if (ls->words) {
        nr = 0;
        for (ws = ls->words; ws; ws = ws->next)
          nr++;
        dprintf(idx, "<center><br><br>%s spoke %d different words today. These are the most used ones:<br>\n", ls->user, nr);
        sortwordstats(ls, NULL);
        dprintf(idx, "<table>\n");
        nr = 0;
        ws = ls->words;
        while (ws && (nr < 10)) {
          nr++;
          dprintf(idx, "<tr><td>%d.</td><td>%s</td><td>(%d)</td></tr>\n", nr, filt2(ws->word), ws->nr);
          ws = ws->next;
        }
        dprintf(idx, "</table><br><br></center>\n");
      }
    }
    dprintf(idx, "<br><br><hr>\n");
    dprintf(idx, "<table width=100%% border=0>\n");
    dprintf(idx, "<tr><td align=center><table width=100%% border=0><tr>\n");
    dprintf(idx, "<td width=25%% align=center><font size=-1><a href=\"../\">other users</a></font></td>\n");
    dprintf(idx, "<td width=25%% align=center><font size=-1><a href=\"../../top/total/words/\">top%d</a></font></td>\n", webnr);
    dprintf(idx, "<td width=25%% align=center><font size=-1><a href=\"../../misc/\">misc stats</a></font></td>\n");
    dprintf(idx, "<td width=25%% align=center><font size=-1><a href=\"/\">other channels</a></font></td>\n");
    dprintf(idx, "</tr></table></td></tr>\n");
    dprintf(idx, "</table>\n");
    dprintf(idx, "<center>Created by <a href=\"http://www.visions-of-fantasy.de/stats.mod/\">Stats.mod v%s</a></center></body></html>", MODULE_VERSION);
    dprintf(idx, "</body></html>");
    return;
  } else if (!strcasecmp(command, "misc")) {
    do_miscstats(idx, channel);
    return;
  } else {
    dprintf(idx, "<html><body>Error! unknown command %s</body></html>", command);
    stats_info_access(idx)->code = 404;
    return;
  }
}

static void do_graphs(int idx, int today, globstats *gs, char *channel)
{
  int itype, nr;
  int total, rest, width, max;
  float onep, percent, wpercent;
  char bground[140];
  locstats *ls;
  struct userrec *u;
  char what[512], *pwhat, *type;


  Context;
  if (graphgif[0] == 0)
    bground[0] = 0;
  else
    sprintf(bground, "background=\"%s\"", graphgif);
  dprintf(idx, "<!-- Created by Stats.mod v%s-->\n", MODULE_VERSION);
  dprintf(idx, "<html><head><meta http-equiv=\"expires\" content=\"0\">\n");
  dprintf(idx, "<title>%s top %d statistics on channel %s</title></head>\n",
          RANGESTR, graphnr, channel);
  dprintf(idx, "%s\n", bodytag);
  dprintf(idx, "<center><h1>%s top %d statistics on channel %s</h1>\n",
          RANGESTR, graphnr, channel);
  sprintf(what, "%s", graphstats);
  pwhat = what;
  while (strlen(pwhat) > 0) {
    type = newsplit(&pwhat);
    itype = typetoi(type);
    if (itype < 0) {
      putlog(LOG_MISC, "*", "Stats.mod: Error serving livestats. Unsupported type %s. Skipping.", type);
      continue;
    }
    sortstats(gs, type, today);
    if (gs->local)
      max = gs->local->values[today][itype];
    else {
      debug1("Error! Can't find local stats for %s", channel);
      dprintf(idx, "<html><body>Error! Can't find local stats for %s</body></html>", channel);
      stats_info_access(idx)->code = 404;
      return;
    }
    total = gettotal(gs, itype, today);
    rest = total;
    dprintf(idx, "<br><br><br><table width=100%% border=1><tr><th width=100%% align=center><font size=+1>Ordered by <b>%s</b></font></td></tr></table>", type);
    dprintf(idx, "<center><font size=-1>[%d %s total]</font></center><br>\n", total, type);
    dprintf(idx, "<table width=100%%>\n");
    if (!total || !max)
      continue;
    onep = (float) total / 100.0;
    nr = 0;
    for (ls = gs->local; ls; ls = ls->next) {
      u = get_user_by_handle(userlist, ls->user);
      if (matchattr(u, nostats, channel))
        continue;
      if (!ls->values[today][itype])
        break;
      nr++;
      if (nr > graphnr)
        break;
      dprintf(idx, "<tr>\n");
      percent = (float) ls->values[today][itype] / onep;
      wpercent = (float) ls->values[today][itype] / ((float) max / 100.0);
      width = (int) wpercent * 0.8;
      dprintf(idx, "  <td align=right width=10%%><a href=\"../../../users/%s/\">%s</a></td>\n", ls->user, ls->user);
      dprintf(idx, "  <td align=left width=90%%>\n");
      dprintf(idx, "    <table width=100%%><tr>\n");
      dprintf(idx, "      <td width=%d%% bgcolor=\"%s\" %s>&nbsp;</td>\n", width, graphcolor, bground);
      dprintf(idx, "      <td align=left width=%d%%>%.2f%% <font size=-2>[%d]</font></td>\n", 100 - width, percent, ls->values[today][itype]);
      dprintf(idx, "    </tr></table>\n");
      dprintf(idx, "  </td>\n");
      dprintf(idx, "</tr>\n");
      rest -= ls->values[today][itype];
    }
    dprintf(idx, "<tr>\n");
    percent = (float) rest / ((float) total / 100.0);
    wpercent = (float) rest / ((float) max / 100.0);
    width = (int) wpercent * 0.8;
    dprintf(idx, "  <td align=right width=10%%>[others]</td>\n");
    dprintf(idx, "  <td align=left width=90%%>\n");
    dprintf(idx, "    <table width=100%%><tr>\n");
    dprintf(idx, "      <td width=%d%% bgcolor=\"%s\" %s>&nbsp;</td>\n", width, graphcolor, bground);
    dprintf(idx, "      <td align=left width=%d%%>%.2f%% <font size=-2>[%d]</font></td>\n", 100 - width, percent, rest);
    dprintf(idx, "    </tr></table>\n");
    dprintf(idx, "  </td>\n");
    dprintf(idx, "</tr></table>\n");
  }
  dprintf(idx, "</table>\n");
  dprintf(idx, "<br><hr><br>\n");
  dprintf(idx, "<table width=100%% border=0>\n");
  dprintf(idx, "<tr><td align=center><table width=100%% border=0><tr>\n");
  dprintf(idx, "<td width=25%% align=center><a href=\"../words/\">table</a></td>\n");
  dprintf(idx, "<td width=25%% align=center><a href=\"../../../misc/\">misc stats</a></td>\n");
  dprintf(idx, "<td width=25%% align=center><a href=\"../../../users/\">users</a></td>\n");
  dprintf(idx, "<td width=25%% align=center><a href=\"/\">other channels</a></td>\n");
  dprintf(idx, "</tr></table></td></tr>\n");
  dprintf(idx, "<tr><td align=center><table width=100%% border=0><tr>\n");
  dprintf(idx, "<td width=25%% align=center><a href=\"../../total/graphs/\">total</a></td>\n");
  dprintf(idx, "<td width=25%% align=center><a href=\"../../daily/graphs/\">daily</a></td>\n");
  dprintf(idx, "<td width=25%% align=center><a href=\"../../weekly/graphs/\">weekly</a></td>\n");
  dprintf(idx, "<td width=25%% align=center><a href=\"../../monthly/graphs/\">monthly</a></td>\n");
  dprintf(idx, "</tr></table></td></tr>\n");
  dprintf(idx, "</table>\n");
  dprintf(idx, "<br><center>Created by <a href=\"http://www.visions-of-fantasy.de/stats.mod/\">Stats.mod v%s</a></center></body></html>", MODULE_VERSION);
  return;
}

static void do_toptalkers(int idx, int today, globstats *gs, char *channel, char *buf)
{
  char what2[512], *pwhat2, *type, *type2, ts[61];
  locstats *ls;
  time_t tt, ttbuf;
  struct userrec *u;
  int itype, itype2, nr;

  Context;
  type = buf;
  itype = typetoi(type);
  if (itype == T_ERROR) {
    dprintf(idx, "<html><body>Error! No such type %s</body></html>", type);
    stats_info_access(idx)->code = 404;
    return;
  }
  dprintf(idx, "<html><head><meta http-equiv=\"expires\" content=\"0\">\n");
  dprintf(idx, "<title>%s top %d talkers on channel %s</title>\n", RANGESTR, webnr, channel);
  dprintf(idx, "<META Name=\"Stats.mod\" Content=\"%s\">\n", MODULE_VERSION);
  dprintf(idx, "<META Name=\"channel\" Content=\"%s\">\n", gs->chan);
  dprintf(idx, "<META Name=\"network\" Content=\"%s\">\n", network);
  dprintf(idx, "<META Name=\"written\" Content=\"%lu\">\n", now);
  dprintf(idx, "</head>\n");
  dprintf(idx, "%s", bodytag);
  dprintf(idx, "<center><h1>%s top %d talkers on channel %s</h1>\n", RANGESTR, webnr, channel);
  dprintf(idx, "Ordered by \"%s\" (to change it click corresponding table field).<BR>\n", type);
  dprintf(idx, "User peak: %d<BR>\n", gs->peak[today]);
  dprintf(idx, "<P><table border=2 cellpadding=1><tr align=right>");
  dprintf(idx, "<th>No</th>");
  dprintf(idx, "<th align=center>Nick</th>");
  sprintf(what2, "%s", webstats);
  pwhat2 = what2;
  while (strlen(pwhat2) > 0) {
    type2 = newsplit(&pwhat2);
    itype2 = typetoi(type2);
    if (itype2 == T_MINUTES)
      dprintf(idx, "<th><a href=\"../%s/\">%s</a></th>\n", type2, "time spent");
    else
      dprintf(idx, "<th><a href=\"../%s/\">%s</a></th>\n", type2, type2);
  }
  sortstats(gs, type, today);
  nr = 0;
  for (ls = gs->local; ls; ls = ls->next) {
    u = get_user_by_handle(userlist, ls->user);
    if (matchattr(u, nostats, channel))
      continue;
    if ((itype >= 0) && !ls->values[today][itype])
      break;
    nr++;
    if (nr > webnr)
      break;
    dprintf(idx, "<tr align=right><td>%d</td><td allign=center><a href=\"../../../users/%s/\">%s</a></td>", nr, ls->user, ls->user);
    sprintf(what2, "%s", webstats);
    pwhat2 = what2;
    while (strlen(pwhat2) > 0) {
      type2 = newsplit(&pwhat2);
      itype2 = typetoi(type2);
      if ((itype2 < 0) && (itype2 != T_WPL))
        continue;
      dprintf(idx, "<td>");
      if (itype == itype2)
        dprintf(idx, "<b>");
      if (itype2 == T_MINUTES)
        dprintf(idx, "%s", duration(ls->values[today][itype2] * 60));
      else if (itype2 == T_WPL) {
        if (ls->values[today][T_LINES])
          dprintf(idx, "%.2f", ((float) ls->values[today][T_WORDS]) / ((float) ls->values[today][T_LINES]));
        else
          dprintf(idx, "0");
      } else if (itype2 >= 0)
        dprintf(idx, "%d", ls->values[today][itype2]);
      else
        dprintf(idx, "ERROR! itype2: %d", itype2);
      if (itype == itype2)
        dprintf(idx, "</b>");
      dprintf(idx, "</td>");
    }
    dprintf(idx, "</tr>\n");
  }
  dprintf(idx, "<tr>");
  dprintf(idx, "<td colspan=%d align=center>", countwords(webstats) + 2);
  dprintf(idx, "<b>Total users: %d", countstatmembers(gs));
  dprintf(idx, "</B></td></tr>");
  /* dprintf(idx, "<tr><td colspan=%d align=center><A HREF=\"../graphs/\">graphs</A></td></tr>\n", countwords(webstats) + 2); */
  dprintf(idx, "<tr><td colspan=%d align=center><table width=100%% border=0><tr>\n", countwords(webstats) + 2);
  dprintf(idx, "<td width=25%% align=center><a href=\"../../total/%s/\">total</a></td>\n", type);
  dprintf(idx, "<td width=25%% align=center><a href=\"../../daily/%s/\">daily</a></td>\n", type);
  dprintf(idx, "<td width=25%% align=center><a href=\"../../weekly/%s/\">weekly</a></td>\n", type);
  dprintf(idx, "<td width=25%% align=center><a href=\"../../monthly/%s/\">monthly</a></td>\n", type);
  dprintf(idx, "</tr></table></td></tr>\n");
  dprintf(idx, "<tr><td colspan=%d align=center><table width=100%% border=0><tr>\n", countwords(webstats) + 2);
  dprintf(idx, "<td width=25%% align=center><a href=\"../graphs/\">graphs</a></td>\n");
  dprintf(idx, "<td width=25%% align=center><a href=\"../../../misc/\">misc stats</a></td>\n");
  dprintf(idx, "<td width=25%% align=center><a href=\"../../../users/\">users</a></td>\n");
  dprintf(idx, "<td width=25%% align=center><a href=\"/\">other channels</a></td>\n");
  dprintf(idx, "</tr></table></td></tr>\n");
  /*
  if (today)
    dprintf(idx, "<tr><td colspan=%d align=center><A HREF=\"../../total/%s/\">total stats</A></td></tr>", countwords(webstats) + 2, type);
  else
    dprintf(idx, "<tr><td colspan=%d align=center><A HREF=\"../../today/%s/\">stats of today</A></td></tr>", countwords(webstats) + 2, type);
  */
  dprintf(idx, "</table><br>");
  tt = now;
  dprintf(idx, "<a href=\"http://www.visions-of-fantasy.de/stats.mod/\">Stats.mod v%s</a>.<BR>", MODULE_VERSION);
  tt = gs->started;
  strftime(ts, 60, "%d.%m. %Y  %H:%M", localtime(&tt));
  if (!today)
    dprintf(idx, "Started %s<BR>", ts);
  ctime(&ttbuf); /* workaround for eggdrop bug */
  dprintf(idx, "</CENTER>");
  dprintf(idx, "</body></html>");
  return;
}

static void do_miscstats(int idx, char *channel)
{
  globstats *gs, *tlds, *isps;
  wordstats *ws, *ws2;
  int nr, nr2;
  locstats *ls;
  char *user, *host, *s;
  topicstr *t;
  char ts[20];
  float f, max, onep;
  int i;
  hoststr *h, *isp, *tld;

  Context;
  gs = findglobstats(channel);
  if (!gs) {
    dprintf(idx, "<html><body>ERROR! Can't find stats for %s!</body></html>\n", channel);
    return;
  }
  dprintf(idx, "<html><head><meta http-equiv=\"expires\" content=\"0\"><title>Misc Stats in %s</title></head>\n", channel);
  dprintf(idx, "%s\n<center><font size=+2>Misc Stats of Today in %s</font><br>\n", bodytag, channel);
  do_globwordstats(gs);
  if (gs->topics) {
    dprintf(idx, "<br><br>Today's Topics:<br>\n<table>");
    for (t = gs->topics; t; t = t->next) {
      strftime(ts, 20, "%H:%M", localtime(&t->when));
      dprintf(idx, "<tr><td>\"%s\"</td><td>(set by %s at %s)</td></tr>\n", filt2(t->topic), t->by, ts);
    }
    dprintf(idx, "</table><br><br>\n");
  }
/*
  gs->users[S_USERSUM][3] = 28;
  gs->users[S_USERCOUNTS][3] = 4;
  gs->users[S_USERSUM][9] = 78;
  gs->users[S_USERCOUNTS][9] = 5;
  gs->users[S_USERSUM][18] = 7;
  gs->users[S_USERCOUNTS][18] = 2;
*/

  max = 0.0;
  for (i = 0; i < 24; i++)
    if (gs->users[S_USERCOUNTS][i])
      if ((((float) gs->users[S_USERSUM][i]) / ((float) gs->users[S_USERCOUNTS][i])) > max)
        max = ((float) gs->users[S_USERSUM][i]) / ((float) gs->users[S_USERCOUNTS][i]);
  if (max > 0.0) {
    dprintf(idx, "Average users:<br><font size=\"-5\"><table border=1 width=\"75%%\">\n");
    onep = max / 100.0;
    for (i = 0; i < 24; i++) {
      if (gs->users[S_USERCOUNTS][i] > 0) {
        f = ((float) gs->users[S_USERSUM][i]) / ((float) gs->users[S_USERCOUNTS][i]);
	dprintf(idx, "<tr><td width=\"0\"><font size=\"-5\">%d.00-%d.59</font></td><td><table width=\"100%%\"><tr>", i, i);
        if (((int) (f / onep)) > 0)
          dprintf(idx, "<td width=\"%d%%\" bgcolor=\"%s\"><font size=\"-5\">&nbsp;</font></td><td><font size=\"-5\">%.2f</font></td></tr></table></td>", (int) (f / onep), graphcolor, f);
        else if ((f / onep) > 0.0)
          dprintf(idx, "<td width=\"100%%\"><font size=\"-5\">%.2f</font></td></tr></table></td>", f);
        else
          dprintf(idx, "<td width=\"100%%\"><font size=\"-5\">%.2f</font></td></tr></table></td>", f);
        dprintf(idx, "</tr>\n");
      }
    }
    dprintf(idx, "</table></font>\n");
  }

  if (gs->hosts) {
    tlds = nmalloc(sizeof(globstats));
    tlds->hosts = NULL;
    isps = nmalloc(sizeof(globstats));
    isps->hosts = NULL;
    for (h = gs->hosts; h; h = h->next) {
      host = strrchr(h->host, '.') + 1;
      if (!atoi(host) && (host[0] != '0')) {
        addhost(host, tlds);
        host = h->host;
        while ((s = strchr(host, '.')) && strchr(s + 1, '.'))
          host = s + 1;
        addhost(host, isps);
      } else {
        addhost("[IP]", tlds);
        addhost("[IP]", isps);
      }
    }
    sorthosts(isps);
    sorthosts(tlds);
    dprintf(idx, "<br><br><table border=1><tr><th colspan=2>Most used</th></tr><tr><th>ISPs</th><th>TLDs</th></tr>\n");
    i = 0;
    isp = isps->hosts;
    tld = tlds->hosts;
    while ((i <= 5) && (isp || tld)) {
      i++;
      dprintf(idx, "<tr>");
      if (isp)
        dprintf(idx, "<td>%s (%d)</td>", isp->host, isp->nr);
      else
        dprintf(idx, "<td>-</td>");
      if (tld)
        dprintf(idx, "<td>%s (%d)</td>", tld->host, tld->nr);
      else
        dprintf(idx, "<td>-</td>");
      dprintf(idx, "</tr>\n");
      if (isp)
        isp = isp->next;
      if (tld)
        tld = tld->next;
    }
    dprintf(idx, "</table>\n");
    free_hosts(isps->hosts);
    free_hosts(tlds->hosts);
    nfree(isps);
    nfree(tlds);
  }





  if (gs->words) {
    nr = 0;
    for (ws = gs->words; ws; ws = ws->next)
      nr++;
    dprintf(idx, "<br>Registered users spoke %d different words<br>\n", nr);
    dprintf(idx, "These are the most used ones:<br><table>\n");
    ws = gs->words;
    nr = 0;
    while (ws && (nr < 10)) {
      nr++;
      dprintf(idx, "<tr><td>%d.</td><td>%s</td><td>(%d)</td></tr>\n", nr, filt2(ws->word), ws->nr);
      ws = ws->next;
    }
    dprintf(idx, "</table>\n");
    nr2 = 0;
    ws2 = NULL;
    user = NULL;
    for (ls = gs->local; ls; ls = ls->next) {
      nr = 0;
      for (ws = ls->words; ws; ws = ws->next)
        nr++;
      if (nr > nr2) {
        nr2 = nr;
        ws2 = ws;
        user = ls->user;
      }
    }
    if (user)
      dprintf(idx, "<br><a href=\"../users/%s/\">%s</a> seems to have the largest vocabulary(%d words).\n",
            user, user, nr2);
  }
  sortstats(gs, "words", S_TODAY);
  if (gs->local && gs->local->values[S_TODAY][T_WORDS]) {
    f = 0.0;
    user = NULL;
    for (ls = gs->local; ls; ls = ls->next) {
      if (ls->values[S_TODAY][T_LINES] && (((float) ls->values[S_TODAY][T_WORDS] / (float) ls->values[S_TODAY][T_LINES]) > f)) {
        f = (float) ls->values[S_TODAY][T_WORDS] / (float) ls->values[S_TODAY][T_LINES];
        user = ls->user;
      }
    }
    if (user)
      dprintf(idx, "<br><a href=\"../users/%s/\">%s</a> likes long speeches and wrote an average of %.2f words per line.\n",
              user, user, f);
  }
  sortstats(gs, "nicks", S_TODAY);
  if (gs->local && (gs->local->values[S_TODAY][T_NICKS] > 1))
    dprintf(idx, "<br><a href=\"../users/%s/\">%s</a> couldn't decide which nick is the best and changed it %d times.\n",
            gs->local->user, gs->local->user, gs->local->values[S_TODAY][T_NICKS]);
  sortstats(gs, "kicks", S_TODAY);
  if (gs->local && gs->local->values[S_TODAY][T_KICKS])
    dprintf(idx, "<br><a href=\"../users/%s/\">%s</a> did what was necessary and kicked %d people.\n",
            gs->local->user, gs->local->user, gs->local->values[S_TODAY][T_KICKS]);
  sortstats(gs, "questions", S_TODAY);
  if (gs->local && gs->local->values[S_TODAY][T_QUESTIONS])
    dprintf(idx, "<br><a href=\"../users/%s/\">%s</a> knows that wisdom is the real wealth and asked %d questions.\n",
            gs->local->user, gs->local->user, gs->local->values[S_TODAY][T_QUESTIONS]);
  sortstats(gs, "joins", S_TODAY);
  if (gs->local && (gs->local->values[S_TODAY][T_JOINS] > 1))
    dprintf(idx, "<br><a href=\"../users/%s/\">%s</a> wasn't too sure whether he/she should stay or not and visited %s %d times.\n",
            gs->local->user, gs->local->user, channel, gs->local->values[S_TODAY][T_JOINS]);
  sortstats(gs, "smileys", S_TODAY);
  if (gs->local && gs->local->values[S_TODAY][T_SMILEYS])
    dprintf(idx, "<br><a href=\"../users/%s/\">%s</a> had a happy day and smiled %d times.\n",
            gs->local->user, gs->local->user, gs->local->values[S_TODAY][T_SMILEYS]);
  dprintf(idx, "<br><br><hr>\n");
  dprintf(idx, "<table width=100%% border=0>\n");
  dprintf(idx, "<tr><td align=center><table width=100%% border=0><tr>\n");
  dprintf(idx, "<td width=33%% align=center><font size=-1><a href=\"../top/total/words/\">top%d</a></font></td>\n", webnr);
  dprintf(idx, "<td width=33%% align=center><font size=-1><a href=\"../users/\">users</a></font></td>\n");
  dprintf(idx, "<td width=33%% align=center><font size=-1><a href=\"/\">other channels</a></font></td>\n");
  dprintf(idx, "</tr></table></td></tr>\n");
  dprintf(idx, "</table>\n");
  dprintf(idx, "<center>Created by <a href=\"http://www.visions-of-fantasy.de/stats.mod/\">Stats.mod v%s</a></center></body></html>", MODULE_VERSION);
  dprintf(idx, "</body></html>");
}