/*
 * Copyright (C) 2000  Florian Sander
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

static int nopubstats(char *chan)
{
  if (!chan) {
    debug0("WARNING: nopubstats() called with NULL pointer");
    return 1;
  }
#if EGG_IS_MIN_VER(10503)
  if (ngetudef("nopubstats", chan))
    return 1;
#endif
  return 0;
}

static int pub_top10(char *nick, char *host, char *hand,
        char *channel, char *text)
{
  putlog(LOG_CMDS, channel, "<<%s>> !%s! top10 %s", nick, hand, text);
  tell_top(channel, channel, text, 1, 10, 0);
  return 1;
}

static int pub_top20(char *nick, char *host, char *hand,
        char *channel, char *text)
{
  putlog(LOG_CMDS, channel, "<<%s>> !%s! top20 %s", nick, hand, text);
  tell_top(channel, channel, text, 11, 20, 0);
  return 1;
}

static int pub_ttop10(char *nick, char *host, char *hand,
        char *channel, char *text)
{
  putlog(LOG_CMDS, channel, "<<%s>> !%s! ttop10 %s", nick, hand, text);
  tell_top(channel, channel, text, 1, 10, 1);
  return 1;
}

static int pub_ttop20(char *nick, char *host, char *hand,
        char *channel, char *text)
{
  putlog(LOG_CMDS, channel, "<<%s>> !%s! ttop20 %s", nick, hand, text);
  tell_top(channel, channel, text, 11, 20, 1);
  return 1;
}

/* static void tell_top(char *channel, char *dest, char *type, int from, int to, int today)
{
  char t[50], result[510], tosend[511], *type2;
  int i, ti;
  struct stats_global *l;
  struct stats_local *ll;
  struct userrec *u;
#if EGG_IS_MIN_VER(10500)
  struct chanset_t *chan;
#endif

  Context;
  filt(type);
  if (nopubstats(channel))
    return;
  if (stat_flood())
    return;
#if EGG_IS_MIN_VER(10500)
  if (!strcasecmp(channel, dest)) {
    chan = findchan_by_dname(dest);
    if (chan)
      dest = chan->name;
  }
#endif
  if (strlen(type) > 100)
    return;
  type2 = newsplit(&type);
  if (!type2 || !type2[0] || (strlen(type2) > 45))
    strcpy(t, "words");
  else
    strcpy(t, type2);
  ti = typetoi(t);
  if ((ti < 0) && (ti != T_WPL) && (ti != T_WORD)) {
    sprintf(tosend, "PRIVMSG %s :No such stat type.\n", dest);
    dprintf(DP_HELP, "%s", tosend);
    return;
  }
  l = findglobstats(channel);
  if (!l) {
    debug1("StatMod: Can't do topx for %s, no such entry in linked list", channel);
    return;
  }
  strlower(type);
  if (ti == T_WORD)
    setword(l, type);
  sortstats(l, t, today);
  sprintf(result, "%sTop %d(\"%s\"):", today ? "Today's " : "", to, (ti == T_WORD) ? type : t);
  i = 1;
  for (ll = l->local; ll; ll = ll->next) {
    if (!ll->word)
      break;
    u = get_user_by_handle(userlist, ll->user);
    if (matchattr(u, nostats, channel))
      continue;
    if ((ti >= 0) && !ll->values[today][ti])
      break;
    if ((i >= from) && (i <= to) && ((strlen(result) + strlen(ll->user) + 10) < 500)) {
      if (ti >= 0)
        sprintf(result, "%s %d. %s(%d)", result, i, ll->user, ll->values[today][ti]);
      else if (ti == T_WPL)
        sprintf(result, "%s %d. %s(%.2f)", result, i, ll->user, (float) ll->values[today][T_WORDS] / (float) ll->values[today][T_LINES]);
      else if (ti == T_WORD)
        sprintf(result, "%s %d. %s(%d)", result, i, ll->user, ll->word->nr);
      else
        sprintf(result, "%s ERROR (%d)", result, ti);
    }
    i++;
  }
  sprintf(tosend, "PRIVMSG %s :%s\n", dest, result);
  dprintf(DP_HELP, "%s", tosend);
  Context;
}
*/

static void tell_top(char *channel, char *dest, char *type, int from, int to, int today)
{
  char t[50], result[510], tosend[511];
  int i, ti;
  struct stats_global *l;
  struct stats_local *ll;
  struct userrec *u;
#if EGG_IS_MIN_VER(10500)
  struct chanset_t *chan;
#endif

  Context;
  filt(type);
  if (stat_flood())
    return;
#if EGG_IS_MIN_VER(10500)
  if (!strcasecmp(channel, dest)) {
    chan = findchan_by_dname(dest);
    if (chan)
      dest = chan->name;
  }
#endif
  if (nopubstats(channel))
    return;
  l = findglobstats(channel);
  if (!l) {
    debug1("StatMod: Can't do topx for %s, no such entry in linked list", channel);
    return;
  }
  if (!strncasecmp(type, "word ", 5)) {
    tell_top_word(channel, dest, type, from, to, l);
    return;
  }
  if (!type[0] || (strlen(type) > 45))
    strcpy(t, "words");
  else
    strcpy(t, type);
  ti = typetoi(t);
  if ((ti < 0) && (ti != T_WPL)) {
    sprintf(tosend, "PRIVMSG %s :No such stat type.\n", dest);
    dprintf(DP_HELP, "%s", tosend);
    return;
  }
  sortstats(l, t, today);
  sprintf(result, "%sTop %d(%s):", today ? "Today's " : "", to, t);
  i = 1;
  for (ll = l->local; ll; ll = ll->next) {
    u = get_user_by_handle(userlist, ll->user);
    if (matchattr(u, nostats, channel))
      continue;
    if ((ti >= 0) && !ll->values[today][ti])
      break;
    if ((i >= from) && (i <= to) && ((strlen(result) + strlen(ll->user) + 10) < 500)) {
      if (ti >= 0)
        sprintf(result, "%s %d. %s(%d)", result, i, ll->user, ll->values[today][ti]);
      else if (ti == T_WPL) {
	if (ll->values[today][T_LINES])
          sprintf(result, "%s %d. %s(%.2f)", result, i, ll->user, (float) ll->values[today][T_WORDS] / (float) ll->values[today][T_LINES]);
        else
          sprintf(result, "%s %d. %s(0)", result, i, ll->user);
      } else
        sprintf(result, "%s ERROR (%d)", result, ti);
    }
    i++;
  }
  sprintf(tosend, "PRIVMSG %s :%s\n", dest, result);
  dprintf(DP_HELP, "%s", tosend);
  Context;
}

static void tell_top_word(char *channel, char *dest, char *type, int from, int to, globstats *gs)
{
  locstats *e;
  struct userrec *u;
  char result[510], tosend[511];
  int i = 1;

  Context;
  newsplit(&type);
  strlower(type);
  setword(gs, type);
  sortstats(gs, "word", 1);
  sprintf(result, "Top %d(\"%s\"):", to, type);
  for (e = gs->local; e; e = e->next) {
    u = get_user_by_handle(userlist, e->user);
    if (matchattr(u, nostats, channel))
      continue;
    if ((i > to) || !e->word)
      break;
    if ((i >= from) && ((strlen(result) + strlen(e->user) + 10) < 500))
      sprintf(result, "%s %d. %s(%d)", result, i, e->user, e->word->nr);
    i++;
  }
  sprintf(tosend, "PRIVMSG %s :%s\n", dest, result);
  dprintf(DP_HELP, "%s", tosend);
  Context;
}

static int pub_place(char *nick, char *host, char *hand,
        char *channel, char *text)
{
  Context;
  putlog(LOG_CMDS, channel, "<<%s>> !%s! place %s", nick, hand, text);
  tell_place(nick, channel, hand, channel, text, S_TOTAL);
  Context;
  return 1;
}

static int pub_tplace(char *nick, char *host, char *hand,
        char *channel, char *text)
{
  Context;
  putlog(LOG_CMDS, channel, "<<%s>> !%s! tplace %s", nick, hand, text);
  tell_place(nick, channel, hand, channel, text, S_TODAY);
  Context;
  return 1;
}

static void tell_place(char *nick, char *dest, char *hand, char *channel, char *text, int today)
{
  globstats *gs;
  locstats *ls;
  char who[20], tosend[512], type[20];
  int place = 0;
  struct userrec *u;
#if EGG_IS_MIN_VER(10500)
  struct chanset_t *chan;
#endif

  Context;
  if (stat_flood())
    return;
#if EGG_IS_MIN_VER(10500)
  if (!strcasecmp(channel, dest)) {
    chan = findchan_by_dname(dest);
    if (chan)
      dest = chan->name;
  }
#endif
  if (nopubstats(channel))
    return;
  if ((text[0] == 0) || (strlen(text) > 19)) {
    strcpy(who, hand);
    strcpy(type, "words");
  } else if (typetoi(text) != T_ERROR) {
    strcpy(who, hand);
    strcpy(type, text);
  } else {
    strcpy(who, text);
    strcpy(type, "words");
  }
  if (who[0] == '*') {
    sprintf(tosend, "PRIVMSG %s :%s, sorry, I don't have any stats about you\n",
            dest, nick);
    dprintf(DP_HELP, "%s", tosend);
    return;
  }
  gs = findglobstats(channel);
  if (!gs) {
    debug1("Stats.mod: Couldn't exec !place, I don't have any statistics in %s.", channel);
    return;
  }
  sortstats(gs, type, today);
  for (ls = gs->local; ls; ls = ls->next) {
    u = get_user_by_handle(userlist, ls->user);
    if (matchattr(u, nostats, channel))
      continue;
    if (!ls->values[today][T_WORDS])
      break;
    place++;
    if (!strcasecmp(who, ls->user))
      break;
  }
  if (!ls) {
    sprintf(tosend, "PRIVMSG %s :%s, sorry, I don't have any stats about %s in %s%s\n",
            dest, nick, who, channel, today ? " today" : "");
    dprintf(DP_HELP, "%s", tosend);
  } else {
    sprintf(tosend, "PRIVMSG %s :%s is nr %d on the list of %stop-talkers.\n", dest,
            who, place, today ? "today's " : "");
    dprintf(DP_HELP, "%s", tosend);
  }
  Context;
}

static int pub_stat(char *nick, char *host, char *hand,
        char *channel, char *text)
{
  Context;
  putlog(LOG_CMDS, channel, "<<%s>> !%s! stat %s", nick, hand, text);
  tell_stat(nick, channel, hand, channel, text, S_TOTAL);
  Context;
  return 1;
}

static int pub_tstat(char *nick, char *host, char *hand,
        char *channel, char *text)
{
  Context;
  putlog(LOG_CMDS, channel, "<<%s>> !%s! tstat %s", nick, hand, text);
  tell_stat(nick, channel, hand, channel, text, S_TODAY);
  Context;
  return 1;
}

static void tell_stat(char *nick, char *dest, char *hand, char *channel, char *text, int today)
{
  locstats *ls;
  char who[20], tosend[512], what[128], *pwhat, *type, *dur;
  int itype;
#if EGG_IS_MIN_VER(10500)
  struct chanset_t *chan;
#endif

  Context;
  if (nopubstats(channel))
    return;
  if (stat_flood())
    return;
#if EGG_IS_MIN_VER(10500)
  if (!strcasecmp(channel, dest)) {
    chan = findchan_by_dname(dest);
    if (chan)
      dest = chan->name;
  }
#endif
  if ((text[0] == 0) || (strlen(text) > 19))
    strcpy(who, hand);
  else
    strcpy(who, text);
  if (who[0] == '*') {
    sprintf(tosend, "PRIVMSG %s :%s, sorry, I don't have any stats about you\n",
            dest, nick);
    dprintf(DP_HELP, "%s", tosend);
    return;
  }
  ls = findlocstats(channel, who);
  if (!ls) {
    sprintf(tosend, "PRIVMSG %s :%s, sorry, I don't have any stats about %s in %s\n",
            dest, nick, who, channel);
    dprintf(DP_HELP, "%s", tosend);
  } else {
    dur = duration(ls->values[today][T_MINUTES] * 60);
    sprintf(tosend, "PRIVMSG %s :%s:", dest, who);
    what[0] = 0;
    pwhat = what;
    strncpy(pwhat, stat_reply, 127);
    pwhat[127] = 0;
    while (strlen(pwhat) > 0) {
      type = newsplit(&pwhat);
      itype = typetoi(type);
      if (itype < 0)
        continue;
      if ((strlen(tosend) + strlen(type) + strlen(channel) + strlen(dur) + 15) < 510) {
        if (itype != T_MINUTES) {
          sprintf(tosend, "%s %d %s", tosend, ls->values[today][itype], type);
        } else {
          sprintf(tosend, "%s %s spent on %s", tosend, dur, channel);
	}
      }
    }
    dprintf(DP_HELP, "%s.\n", tosend);
  }
  Context;
}

static int pub_wordstats(char *nick, char *host, char *hand,
        char *channel, char *text)
{
  Context;
  putlog(LOG_CMDS, channel, "<<%s>> !%s! wordstats %s", nick, hand, text);
  tell_wordstats(nick, channel, hand, channel, text);
  Context;
  return 0;
}

static void tell_wordstats(char *nick, char *dest, char *hand, char *channel, char *text)
{
  locstats *ls;
  wordstats *ws;
  char who[20], tosend[512];
  int i;
#if EGG_IS_MIN_VER(10500)
  struct chanset_t *chan;
#endif


  Context;
  if (nopubstats(channel))
    return;
  filt(text);
  if (stat_flood())
    return;
#if EGG_IS_MIN_VER(10500)
  if (!strcasecmp(channel, dest)) {
    chan = findchan_by_dname(dest);
    if (chan)
      dest = chan->name;
  }
#endif
  if ((text[0] == 0) || (strlen(text) > 19))
    strcpy(who, hand);
  else
    strcpy(who, text);
  if (who[0] == '*') {
    sprintf(tosend, "PRIVMSG %s :%s, sorry, I don't have any stats about you\n",
            dest, nick);
    dprintf(DP_HELP, "%s", tosend);
    return;
  }
  ls = findlocstats(channel, who);
  if (!ls) {
    sprintf(tosend, "PRIVMSG %s :%s, sorry, I don't have any stats about %s in %s\n",
            dest, nick, who, channel);
    dprintf(DP_HELP, "%s", tosend);
  } else {
    if (!ls->words) {
      sprintf(tosend, "PRIVMSG %s :%s, sorry, I don't have any wordstats about %s in %s\n",
              dest, nick, who, channel);
      dprintf(DP_HELP, "%s", tosend);
    } else {
      sprintf(tosend, "PRIVMSG %s :words most used by %s:", dest, who);
      i = 0;
      sortwordstats(ls, NULL);
      ws = ls->words;
      while (ws && ((strlen(tosend) + strlen(ws->word) + 10) < 450) && (i < 10)) {
        i++;
        sprintf(tosend, "%s %d. %s (%d)", tosend, i, ws->word, ws->nr);
        ws = ws->next;
      }
      dprintf(DP_HELP, "%s\n", tosend);
    }
  }
  Context;
}

static int pub_topwords(char *nick, char *host, char *hand,
        char *channel, char *text)
{
  Context;
  putlog(LOG_CMDS, channel, "<<%s>> !%s! topwords %s", nick, hand, text);
  tell_topwords(nick, channel, hand, channel);
  Context;
  return 0;
}

static void tell_topwords(char *nick, char *dest, char *hand, char *channel)
{
  globstats *gs;
  wordstats *ws;
  char tosend[512];
  int i;
#if EGG_IS_MIN_VER(10500)
  struct chanset_t *chan;
#endif

  Context;
  if (nopubstats(channel))
    return;
  if (stat_flood())
    return;
#if EGG_IS_MIN_VER(10500)
  if (!strcasecmp(channel, dest)) {
    chan = findchan_by_dname(dest);
    if (chan)
      dest = chan->name;
  }
#endif
  gs = findglobstats(channel);
  if (!gs) {
    dprintf(DP_HELP, "PRIVMSG %s :%s, I don't have any stats in %s\n", dest, nick, channel);
    return;
  }
  do_globwordstats(gs);
  ws = gs->words;
  if (!ws) {
    dprintf(DP_HELP, "PRIVMSG %s: I don't have any wordstats in %s\n", dest, channel);
    return;
  }
  sprintf(tosend, "PRIVMSG %s :Most used words in %s:", dest, channel);
  i = 0;
  while (ws && ((strlen(tosend) + strlen(ws->word) + 10) < 450) && (i < 10)) {
    i++;
    sprintf(tosend, "%s %d. %s (%d)", tosend, i, ws->word, ws->nr);
    ws = ws->next;
  }
  dprintf(DP_HELP, "%s\n", tosend);
  Context;
}

static int stat_flood()
{
  if (!maxstat_thr || !maxstat_time)
    return 0;
  if ((now - mstat_time) > maxstat_time) {
    mstat_time = now;
    mstat_thr = 0;
  }
  mstat_thr++;
  if (mstat_thr > maxstat_thr)
    return 1;
  return 0;
}

static cmd_t stats_pub[] =
{
  {"!top10", "", pub_top10, 0},
  {"!ttop10", "", pub_ttop10, 0},
  {"!top20", "", pub_top20, 0},
  {"!ttop20", "", pub_ttop20, 0},
  {"!place", "", pub_place, 0},
  {"!tplace", "", pub_tplace, 0},
  {"!stat", "", pub_stat, 0},
  {"!tstat", "", pub_tstat, 0},
  {"!wordstats", "", pub_wordstats, 0},
  {"!topwords", "", pub_topwords, 0},
  {0, 0, 0, 0}
};
